"""PSUtils libpaper interface.

Copyright (c) Reuben Thomas 2023-2025.
Released under the GPL version 3, or (at your option) any later version.
"""

import os
import re
import subprocess

from .types import Rectangle
from .warnings import die


# Get the size of the given paper, or the default paper if no argument given.
def paper(cmd: list[str], silent: bool = False) -> str | None:
    cmd.insert(0, "paper")
    try:
        out = subprocess.check_output(
            cmd,
            stderr=subprocess.DEVNULL if silent else None,
            text=True,
            env=dict(os.environ, LC_NUMERIC="C"),
        )
        return out.rstrip()
    except subprocess.CalledProcessError:
        return None
    except Exception:
        die("could not run `paper' command")


def get_paper_size(paper_name: str | None = None) -> Rectangle | None:
    if paper_name is None:
        paper_name = paper(["--no-size"])
    dimensions: str | None = None
    if paper_name is not None:
        dimensions = paper(["--unit=pt", paper_name], True)
    if dimensions is None:
        return None
    m = re.search(" ([.0-9]+)x([.0-9]+) pt$", dimensions)
    assert m
    w, h = float(m[1]), float(m[2])
    return Rectangle(round(w), round(h))  # round dimensions to nearest point



"""pstops command.

Copyright (c) Reuben Thomas 2023-2025.
Released under the GPL version 3, or (at your option) any later version.
"""

import argparse
import sys
import warnings
from typing import NoReturn

from psutils.argparse import (
    HelpFormatter,
    PaperContext,
    add_draw_argument,
    add_file_arguments,
    add_paper_arguments,
    add_quiet_and_help_arguments,
    add_version_argument,
    parserange,
    parsespecs,
)
from psutils.transformers import file_transform
from psutils.types import Rectangle
from psutils.warnings import simple_warning


DEFAULT_SPECS = "0"


# Command-line parsing helper functions
def get_parser() -> tuple[argparse.ArgumentParser, PaperContext]:
    # Command-line arguments
    parser = argparse.ArgumentParser(
        description="Rearrange pages of a PDF or PostScript document.",
        formatter_class=HelpFormatter,
        usage="%(prog)s [OPTION...] [INFILE [OUTFILE]]",
        add_help=False,
        epilog=f"""
PAGES is a comma-separated list of pages and page ranges.

SPECS is a list of page specifications [default is "{DEFAULT_SPECS}", which selects
each page in its normal order].
""",
    )
    warnings.showwarning = simple_warning(parser.prog)
    paper_context = PaperContext()

    # Command-line parser
    parser.add_argument(
        "-S",
        "--specs",
        help="page specifications (see below)",
    )
    parser.add_argument(
        "-R",
        "--pages",
        dest="pagerange",
        type=parserange,
        help="select the given page ranges",
    )
    parser.add_argument(
        "-e",
        "--even",
        action="store_true",
        help="select even-numbered output pages",
    )
    parser.add_argument(
        "-o",
        "--odd",
        action="store_true",
        help="select odd-numbered output pages",
    )
    parser.add_argument(
        "-r",
        "--reverse",
        action="store_true",
        help="reverse the order of the output pages",
    )
    add_paper_arguments(parser)
    add_draw_argument(parser, paper_context)
    parser.add_argument("-b", "--nobind", help=argparse.SUPPRESS)
    add_version_argument(parser)
    add_quiet_and_help_arguments(parser)
    add_file_arguments(parser)
    # Hidden argument for backwards compatibility.
    parser.add_argument("specs_alt", nargs="?", help=argparse.SUPPRESS)

    return parser, paper_context


def get_parser_manpages() -> argparse.ArgumentParser:
    return get_parser()[0]


class SpecsException(Exception):
    pass


def spec_exception() -> NoReturn:
    raise SpecsException("invalid specs")


def pstops(argv: list[str] = sys.argv[1:]) -> None:
    parser, paper_context = get_parser()
    args = parser.parse_intermixed_args(argv)

    # Get specs if we don't have them yet
    if args.specs is None:
        if args.infile is None:
            args.specs = "-"
        else:
            args.specs = args.infile
        try:
            parsespecs(args.specs, paper_context, err_function=spec_exception)
            args.infile = args.outfile
            args.outfile = args.specs_alt
        except SpecsException:
            args.specs = DEFAULT_SPECS

    size: Rectangle | None = None
    in_size: Rectangle | None = None
    if args.paper:
        size = args.paper
    elif args.width is not None and args.height is not None:
        size = Rectangle(args.width, args.height)
    if args.inpaper:
        in_size = args.inpaper
    elif args.inwidth is not None and args.inheight is not None:
        in_size = Rectangle(args.inwidth, args.inheight)
    if size:
        paper_context = PaperContext(size)
    specs, modulo, flipping = parsespecs(args.specs, paper_context)

    with file_transform(
        args.infile,
        args.outfile,
        size,
        in_size,
        specs,
        args.draw,
        False,
    ) as transform:
        transform.transform_pages(
            args.pagerange,
            flipping,
            args.reverse,
            args.odd,
            args.even,
            modulo,
            args.verbose,
        )


if __name__ == "__main__":
    pstops()



"""psselect command.

Copyright (c) Reuben Thomas 2023.
Released under the GPL version 3, or (at your option) any later version.
"""

import argparse
import sys
import warnings

from psutils.argparse import (
    HelpFormatter,
    PaperContext,
    add_basic_arguments,
    parserange,
    parsespecs,
)
from psutils.transformers import file_transform
from psutils.warnings import die, simple_warning


def get_parser() -> argparse.ArgumentParser:
    # Command-line arguments
    parser = argparse.ArgumentParser(
        description="Select pages from a PDF or PostScript document.",
        formatter_class=HelpFormatter,
        usage="%(prog)s [OPTION...] [PAGES] [INFILE [OUTFILE]]",
        add_help=False,
        epilog="""
PAGES is a comma-separated list of pages and page ranges; see
pstops(1) for more details.
""",
    )
    warnings.showwarning = simple_warning(parser.prog)

    # Command-line parser
    parser.add_argument("-R", "-p", "--pages", help="select the given page ranges")
    parser.add_argument(
        "-e", "--even", action="store_true", help="select even-numbered output pages"
    )
    parser.add_argument(
        "-o", "--odd", action="store_true", help="select odd-numbered output pages"
    )
    parser.add_argument(
        "-r",
        "--reverse",
        action="store_true",
        help="reverse the order of the output pages",
    )
    parser.add_argument("alt_pages", metavar="PAGES", nargs="?", help=argparse.SUPPRESS)
    add_basic_arguments(parser)

    return parser


def psselect(argv: list[str] = sys.argv[1:]) -> None:
    args = get_parser().parse_intermixed_args(argv)

    # Get page range argument if supplied as a non-option
    if args.alt_pages is not None:
        if (
            args.pages is None
            and args.even is False
            and args.odd is False
            and args.reverse is False
        ):
            args.pages = args.alt_pages
        else:
            if args.outfile is not None:
                if args.pages is not None:
                    die("PAGES specified both with and without an option flag")
                else:
                    die("--pages must be used when --even, --odd or --reverse is used")
            args.outfile = args.infile
            args.infile = args.alt_pages

    # Rearrange pages
    pagerange = parserange(args.pages) if args.pages is not None else None
    paper_context = PaperContext()
    specs, modulo, flipping = parsespecs("0", paper_context)
    with file_transform(
        args.infile, args.outfile, None, None, specs, 0, False
    ) as transform:
        transform.transform_pages(
            pagerange, flipping, args.reverse, args.odd, args.even, modulo, args.verbose
        )


if __name__ == "__main__":
    psselect()



"""PSUtils I/O utilities.

Copyright (c) Reuben Thomas 2023-2025.
Released under the GPL version 3, or (at your option) any later version.
"""

import io
import os
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from typing import IO

import puremagic

from .warnings import die


@contextmanager
def setup_inputs_and_output(
    infile_names: list[str | None], outfile_name: str | None
) -> Iterator[tuple[list[tuple[IO[bytes], str]], IO[bytes]]]:
    # Set up inputs
    seekable_infiles: list[tuple[IO[bytes], str]] = []
    for infile_name in infile_names:
        infile: IO[bytes]
        if infile_name is None or infile_name == "-":
            infile = os.fdopen(sys.stdin.fileno(), "rb", closefd=False)
        else:
            try:
                infile = open(infile_name, "rb")
            except OSError:
                die(f"cannot open input file {infile_name}")

        # Find MIME type of input
        data = infile.read(16)
        file_type = puremagic.from_string(data)

        # Slurp infile into a seekable BytesIO
        seekable_infile = io.BytesIO(data + infile.read())
        infile.close()

        seekable_infiles.append((seekable_infile, file_type))

    # Set up output
    if outfile_name is None or outfile_name == "-":
        outfile = os.fdopen(sys.stdout.fileno(), "wb", closefd=False)
    else:
        try:
            outfile = open(outfile_name, "wb")
        except OSError:
            die(f"cannot open output file {outfile_name}")

    # Context manager
    try:
        yield seekable_infiles, outfile
    finally:
        for infile, _ in seekable_infiles:
            infile.close()
        outfile.close()


@contextmanager
def setup_input_and_output(
    infile_name: str | None, outfile_name: str | None
) -> Iterator[tuple[IO[bytes], str, IO[bytes]]]:
    with setup_inputs_and_output([infile_name], outfile_name) as (infiles, outfile):
        yield infiles[0][0], infiles[0][1], outfile
